<?php
declare(strict_types=1);

namespace Inti\Wp\Provider\Plugin;

use Doctrine\Common\Collections\ArrayCollection;
use Doctrine\Common\Collections\Collection;
use Inti\Provider\Service\ProviderManagerInterface;
use Inti\Wp\Plugin\PluginInterface;
use Inti\Wp\Provider\Metadata\MetadataProviderInterface;
use Inti\Wp\Provider\Service\IdManager;
use Inti\Wp\Provider\Type\Post\PostTypeInterface;
use Inti\Wp\Provider\Type\Term\TaxonomyInterface;
use Laminas\Stdlib\ArrayUtils;
use WP_Term;

use function add_filter;
use function array_map;
use function get_post_type;
use function get_term;
use function in_array;
use function is_array;

/**
 * Class ProvideMetadata
 *
 * @package Inti\Wp\Provider\Plugin
 * @author Danijel Fabijan <hipnaba@gmail.com>
 * @link https://bitbucket.org/oktan-nautika/inti-wp-provider
 */
final class ProvideMetadata implements PluginInterface
{
    private ProviderManagerInterface $providers;
    private IdManager $ids;

    public function __construct(ProviderManagerInterface $providers, IdManager $idOffsets)
    {
        $this->providers = $providers;
        $this->ids = $idOffsets;
    }

    public function run(?array $options = null): void
    {
        $registered = [];
        $providers = $this->providers->getProviders();

        foreach ($providers as $provider) {
            if (!$provider instanceof MetadataProviderInterface) {
                continue;
            }

            $providedType = $provider->getProvidedType();
            $meta_type = $providedType instanceof PostTypeInterface
                ? 'post'
                : ($providedType instanceof TaxonomyInterface ? 'term' : '');
            $action = 'get_' . $meta_type . '_metadata';

            if (!in_array($action, $registered)) {
                add_filter($action, function ($metadata, $object_id, $meta_key, $single) use ($meta_type) {
                    return $this->get_metadata($metadata, $object_id, $meta_key, $single, $meta_type);
                }, 10, 4);
                $registered[] = $action;
            }
        }
    }

    /** @noinspection PhpUnusedParameterInspection */
    public function get_metadata(?array $metadata, int $object_id, string $meta_key, bool $single, string $meta_type): ?array
    {
        /** @var MetadataProviderInterface[]|Collection $providers */
        $providers = $this->providers->getProviders();

        switch ($meta_type) {
            case 'post':
                $type = get_post_type($object_id);
                if (false === $type) return $metadata;

                $providers = (new ArrayCollection($providers))->filter(function (MetadataProviderInterface $provider) use ($type, $meta_key) {
                    $typeMatches = $provider->getProvidedType()->getName() === $type;

                    if (empty($meta_key)) {
                        return $typeMatches;
                    }

                    return $typeMatches && $provider->hasMetadataKey($meta_key);
                });
                break;
            case 'term':
                /** @var WP_Term $term */
                $term = get_term($object_id);

                $providers = (new ArrayCollection($providers))->filter(function (MetadataProviderInterface $provider) use ($term, $meta_key) {
                    $typeMatches = $provider->getProvidedType()->getName() === $term->taxonomy;

                    if (empty($meta_key)) {
                        return $typeMatches;
                    }

                    return $typeMatches && $provider->hasMetadataKey($meta_key);
                });
                break;
            default:
                return $metadata;
        }

        if (count($providers) === 0) {
            return $metadata;
        }

        if ($this->ids->isOffset($object_id)) {
            $object_id = $this->ids->unoffset($object_id);
        }

        foreach ($providers as $provider) {
            $providedMetadata = $provider->getMetadata($object_id);
            $providedMetadata = array_map(function ($value) {
                return !is_array($value) ? [$value] : $value;
            }, $providedMetadata);

            $metadata = ArrayUtils::merge($metadata ?? [], $providedMetadata);
        }

        return !empty($meta_key)
            ? $metadata[$meta_key] ?? null
            : $metadata;
    }


}
