<?php
declare(strict_types=1);

namespace Inti\Wp\Provider\Plugin;

use DateTime;
use DateTimeZone;
use Inti\Provider\Query\QueryInterface;
use Inti\Provider\Service\ProviderManagerInterface;
use Inti\Provider\Type\TypeInterface;
use Inti\Wp\Plugin\PluginInterface;
use Inti\Wp\Provider\Post\ProviderInterface;
use Inti\Wp\Provider\Service\IdManager;
use Inti\Wp\Provider\Service\SqlHelpers;
use Inti\Wp\Provider\Type\Post\PostTypeInterface;
use stdClass;
use WP_Query;

use function add_action;
use function add_filter;
use function array_map;
use function get_taxonomy;
use function get_term_by;
use function intval;
use function is_admin;
use function is_array;
use function is_numeric;
use function is_string;
use function max;
use function preg_match;
use function register_post_type;
use function wp_cache_set;
use function wp_dropdown_categories;

/**
 * Class ProvidePosts
 *
 * @package Inti\Wp\Provider\Plugin
 * @author Danijel Fabijan <hipnaba@gmail.com>
 * @link https://bitbucket.org/oktan-nautika/inti-wp-provider
 */
final class ProvidePosts implements PluginInterface
{
    use SqlHelpers;

    private ProviderManagerInterface $providers;
    private IdManager $ids;
    /** @var array<string, TypeInterface> */
    private array $provided = [];

    public function __construct(ProviderManagerInterface $providers, IdManager $idOffsets)
    {
        $this->providers = $providers;
        $this->ids = $idOffsets;
    }

    public function run(?array $options = null): void
    {
        $providers = $this->providers->getProviders();

        if (count($providers) === 0) {
            return;
        }

        add_action('init', [$this, 'register_post_types']);
        add_filter('posts_pre_query', [$this, 'posts_pre_query'], 10, 2);
        add_filter('query', [$this, 'query']);
        add_filter('wp_count_posts', [$this, 'wp_count_posts'], 10, 2);

        foreach ($providers as $provider) {
            $type = $provider->getProvidedType();
            $this->provided[$type->getName()] = $type;
            wp_cache_set('posts-' . $type->getName(), new stdClass(), 'counts');
        }
    }

    public function posts_pre_query($posts, WP_Query $query): ?array
    {
        $q = &$query->query_vars;

        if (!is_string($q['post_type']) || empty($q['post_type']) || !isset($this->provided[$q['post_type']])) {
            return $posts;
        }

        $provider = $this->providers->getProvider($this->provided[$q['post_type']]);

        if ($provider instanceof ProviderInterface) {
            $provider->query();

            $limit = $q['posts_per_page'];
            $page = max(1, intval($q['paged']));
            $offset = ($page - 1) * $limit;

            $provider->limit($limit)->offset($offset);

            if (!empty($q['p']) && is_numeric($q['p'])) {
                $id = intval($q['p']);
                $provider->id(
                    $this->ids->unoffset($id)
                );
            }

            if (is_array($q['post__in']) && !empty($q['post__in'])) {
                $ids = array_map('intval', $q['post__in']);
                $provider->ids(
                    array_map([$this->ids, 'unoffset'], $ids)
                );
            }

            if (!empty($q['s'])) {
                $provider->search($q['s']);
            }

            if (!empty($q['name'])) {
                $provider->slug($q['name']);
            }

            foreach ($provider->getItems() as $row) {
                $posts[] = (object) $this->prepareRow($row, $provider->getProvidedType());
            }

            $query->found_posts = $provider->getCount();
            $query->max_num_pages = ceil($query->found_posts / $limit);
        }

        return $posts;
    }

    public function query(string $query): string
    {
        if (preg_match('/^SELECT \* FROM wp_posts WHERE ID = (\d+) LIMIT 1$/', $query, $matches)) {
            $post_id = intval($matches[1]);

            if ($this->ids->isOffset($post_id)) {
                $type = $this->ids->getType($post_id);
                $id = $this->ids->unoffset($post_id);

                if (($provider = $this->providers->getProvider($type)) && $provider instanceof QueryInterface) {
                    $provider->query()
                        ->id($id)
                        ->limit(1);

                    if ($row = $provider->getItem()) {
                        $row = $this->prepareRow($row, $type);
                        $query = $this->arrayToSelect($row);
                    }
                }
            }
        }

        return $query;
    }

    public function register_post_types(): void
    {
        foreach ($this->providers->getProviders() as $provider) {
            $type = $provider->getProvidedType();

            if ($type instanceof PostTypeInterface && !$type->isBuiltin()) {
                $this->ids->registerOffset($type);

                register_post_type($type->getName(), [
                    'description' => $type->getDescription(),
                    'hierarchical' => $type->isHierarchical(),
                    'label' => $type->getLabel(),
                    'labels' => $type->getLabels(),
                    'public' => $type->isPublic(),
                    'publicly_queryable' => $type->isPubliclyQueryable(),
                    'show_ui' => $type->showUi(),
                    'show_in_menu' => $type->showInMenu(),
                    'show_in_nav_menus' => $type->showInNavMenus(),
                    'show_in_admin_bar' => $type->showInAdminBar(),
                    'menu_position' => $type->getMenuPosition(),
                    'menu_icon' => $type->getMenuIcon(),
                    'capability_type' => $type->getCapabilityType(),
                    'capabilities' => $type->getCapabilities(),
                    'map_meta_cap' => $type->mapMetaCap(),
                    'supports' => $type->getSupportedFeatures(),
                    'register_meta_box_cb' => $type->getMetaBoxCallback(),
                    'taxonomies' => $type->getTaxonomies(),
                    'has_archive' => $type->hasArchive(),
                    'rewrite' => $type->getRewriteConfig(),
                    'query_var' => $type->getQueryVar(),
                    'can_export' => false,
                    'delete_with_user' => false,
                    'show_in_rest' => false,
                    'rest_base' => false,
                    'rest_controller_class' => false,
                    'template' => [],
                    'template_lock' => false,
                ]);

                if (is_admin()) {
                    add_action('restrict_manage_posts', function() use ($type) {
                        global $typenow;

                        if ($typenow !== $type->getName()) {
                            return;
                        }

                        foreach ($type->getTaxonomies() as $taxonomy) {
                            $this->taxonomy_filter($taxonomy);
                        }
                    });

                    add_action('parse_query', function (WP_Query $query) use ($type) {
                        global $pagenow;

                        if ($pagenow !== 'edit.php') {
                            return;
                        }

                        $vars = &$query->query_vars;

                        if ($vars['post_type'] !== $type->getName()) {
                            return;
                        }

                        foreach ($type->getTaxonomies() as $taxonomy) {
                            if (!isset($vars[$taxonomy]) || !is_numeric($vars[$taxonomy]) || $vars[$taxonomy] === 0) {
                                continue;
                            }

                            $term = get_term_by('id', $vars[$taxonomy], $taxonomy);
                            $vars[$taxonomy] = $term->slug;
                        }
                    });
                }
            }
        }
    }

    public function taxonomy_filter(string $taxonomy): void
    {
        $selected = $_GET[$taxonomy] ?? '';
        $tax_obj = get_taxonomy($taxonomy);

        wp_dropdown_categories([
            'show_option_all' => 'Prikaži sve ' . $tax_obj->label,
            'taxonomy' => $taxonomy,
            'name' => $taxonomy,
            'orderby' => 'name',
            'selected' => $selected,
            'show_count' => true,
            'hide_empty' => false,
        ]);
    }

    public function wp_count_posts(object $count, string $type): object
    {
        $providers = $this->providers->getProviders();

        foreach ($providers as $provider) {
            if ($provider instanceof QueryInterface && $provider->getProvidedType()->getName() === $type) {
                $count->publish = $provider->getCount();
            }
        }

        return $count;
    }

    private function prepareRow(array $row, TypeInterface $type): array
    {
        $row['ID'] = $this->ids->offset(intval($row['ID']), $type);
        $row['post_author'] = 0;
        $row['post_date'] = $row['post_date'] ?? (new DateTime('now'))->format('Y-m-d H:i:s');
        /** @noinspection PhpUnhandledExceptionInspection */
        $row['post_date_gmt'] = $row['post_date_gmt'] ?? (new DateTime('now', new DateTimeZone('UTC')))->format('Y-m-d H:i:s');
        $row['post_content'] = $row['post_content'] ?? '';
        $row['post_title'] = $row['post_title'] ?? '';
        $row['post_excerpt'] = $row['post_excerpt'] ?? '';
        $row['post_status'] = $row['post_status'] ?? 'publish';
        $row['comment_status'] = 'closed';
        $row['ping_status'] = 'closed';
        $row['post_password'] = '';

        $row['post_type'] = $type->getName();
        $row['post_mime_type'] = '';
        $row['comment_count'] = 0;

        return $row;
    }
}
